/* devSTR7201.c */

/* devSTR7201.c - Device Support Routines for Struck STR7201 multiscaler
 * for EPICS MCA record
 *
 *      Author:  Mark Rivers
 *      Date:    11/9/97
 *
 * Modification Log:
 * -----------------
 * .01  11/09/97  mlr  Developed from AIM device support
 * .02  08/05/99  tmm  Call sendChannelAdvance from initRecord
 */


#include        <vxWorks.h>
#include        <types.h>
#include        <stdioLib.h>
#include        <wdLib.h>
#include        <memLib.h>
#include        <string.h>

#include        <alarm.h>
#include        <callback.h>
#include        <dbDefs.h>
#include        <dbAccess.h>
#include        <recSup.h>
#include        <devSup.h>
#include        <link.h>

#include        <mcaRecord.h>
#include        <mca.h>
#include        "drvSTR7201.h"

#define Debug(l,FMT,V) {if (l <= devSTR7201Debug) \
                          { printf("%s(%d):",__FILE__,__LINE__); \
                            printf(FMT,V);}}
volatile int devSTR7201Debug = 0;

#define MAX(a, b) ((a) > (b) ? (a) : (b) )

/* Create DSET */
static long init_record();
static long send_msg();
static long read_array();
static long sendPresets(mcaRecord *pmca);
static long sendChannelAdvance(mcaRecord *pmca);

typedef struct {
        long            number;
        DEVSUPFUN       report;
        DEVSUPFUN       init;
        DEVSUPFUN       init_record;
        DEVSUPFUN       get_ioint_info;
        DEVSUPFUN       send_msg;
        DEVSUPFUN       read_array;
} STR7201_DSET;

STR7201_DSET devSTR7201 = {
        6,
        NULL,
        NULL,
        init_record,
        NULL,
        send_msg,
        read_array
};

static long init_record(mcaRecord *pmca)
{
    int card;
    Debug(5, "devSTR7201(init_record): entry\n", 0);
    /* pmca->inp must be VME_IO */
    switch (pmca->inp.type) {
    case (VME_IO) :
        card = pmca->inp.value.vmeio.card;
        if (drvSTR7201Init(card) != OK) {
            recGblRecordError(S_db_badField,(void *)pmca,
                "devSTR7201 (init_record) Card not found");
            return(S_db_badField);
        }
        pmca->nord = 0;
        break;
    default :
        recGblRecordError(S_db_badField,(void *)pmca,
                "devSTR7201 (init_record) Illegal INP field");
        return(S_db_badField);
    }
    drvSTR7201SetNChans(card, pmca->nuse);
    sendPresets(pmca);
    sendChannelAdvance(pmca);
    return(0);
}

static long send_msg(mcaRecord *pmca, unsigned long msg, void *parg)
{
        int card = pmca->inp.value.vmeio.card;
        int signal = pmca->inp.value.vmeio.signal;
        long s, acq_status, ecounts;
        float ereal;

        switch (msg) {
        case MSG_ACQUIRE:
                /* start acquisition */
                s = drvSTR7201AcqOn(card);
                Debug(5, "devSTR7201(send_msg): start acquisition %d\n", s);
                break;
        case MSG_READ:
                /* start read operation */
                /* This is a no-op. Read-array does everything. */
                break;
        case MSG_SET_CHAS_INT:
                /* set channel advance source to internal (timed) */
                sendChannelAdvance(pmca);
                break;
        case MSG_SET_CHAS_EXT:
                /* set channel advance source to external */
                sendChannelAdvance(pmca);
                break;
        case MSG_SET_NCHAN:
                /* set number of channels */
                s = drvSTR7201SetNChans(card, *(long *)parg);
                Debug(5, "devSTR7201(send_msg): acqu setup %d\n", s);
                break;
        case MSG_SET_DWELL:
                /* set dwell time per channel. */
                sendChannelAdvance(pmca);
                break;
        case MSG_SET_PSCL:
                /* set channel advance prescaler. */
                sendChannelAdvance(pmca);
                break;
        case MSG_SET_REAL_TIME:
                /* set preset real time. */
                sendPresets(pmca);
                break;
        case MSG_SET_LIVE_TIME:
                /* set preset live time */
                sendPresets(pmca);
                break;
        case MSG_SET_COUNTS:
                /* set preset counts */
                sendPresets(pmca);
                break;
        case MSG_SET_LO_CHAN:
                /* set preset count low channel */
                sendPresets(pmca);
                break;
        case MSG_SET_HI_CHAN:
                /* set preset count high channel */
                sendPresets(pmca);
                break;
        case MSG_SET_NSWEEPS:
                /* set number of sweeps (for MCS mode)
                 * This is a NOOP on STR7201 but should be implemented in
                 * software in the future
                 */
                break;
        case MSG_SET_MODE_PHA:
                /* set mode to pulse height analysis */
                /* This is a NOOP for STR7201 */
                break;
        case MSG_SET_MODE_MCS:
                /* set mode to MultiChannel Scaler */
                /* This is a NOOP for STR7201 */
                break;
        case MSG_SET_MODE_LIST:
                /* set mode to LIST (record each incoming event) */
                /* This is a NOOP for STR7201 */
                break;
        case MSG_GET_ACQ_STATUS:
                /* Read the current status of the device */
                s = drvSTR7201GetAcqStatus(card, signal, &ereal, &ecounts, 
                                                &acq_status);
                Debug(5, "devSTR7201(send_msg): status update %d\n", s);
                pmca->ertm = ereal;
                pmca->eltm = ereal;
                pmca->act = ecounts;
                pmca->acqg = acq_status;
                break;
        case MSG_STOP_ACQUISITION:
                /* stop data acquisition */
                s = drvSTR7201AcqOff(card);
                Debug(5, "devSTR7201(send_msg): stop acquisition %d\n", s);
                break;
        case MSG_ERASE:
                /* erase */
                s = drvSTR7201Erase(card);
                Debug(5, "devSTR7201(send_msg): erase %d\n", s);
                break;
        }
        return(0);
}

static long read_array(mcaRecord *pmca)
{
        int card = pmca->inp.value.vmeio.card;
        int signal = pmca->inp.value.vmeio.signal;
        int s, nuse = pmca->nuse;
        long *lptr = (long *)pmca->bptr;

        s = drvSTR7201Read(card, signal, nuse, lptr);
        Debug(5, "devSTR7201(read_array): %d\n", s);
        pmca->nord = nuse;
        return(0);
}

static long sendPresets(mcaRecord *pmca)
{
    int card = pmca->inp.value.vmeio.card;
    float ptime;

    ptime = MAX(pmca->prtm, pmca->pltm);
    drvSTR7201SetPresets(card, ptime, pmca->pctl, pmca->pcth, pmca->pct);
    return (OK);
}

static long sendChannelAdvance(mcaRecord *pmca)
{
    int card = pmca->inp.value.vmeio.card;
    int source = LNE_EXTERNAL;
    int prescale = 0;

    if (pmca->chas == mcaCHAS_Internal) source = LNE_INTERNAL;
    if (source == LNE_INTERNAL) {
        prescale = NINT(pmca->dwel * INTERNAL_CLOCK) - 1;
    } else {
        prescale = pmca->pscl - 1;
    }
    drvSTR7201SetChannelAdvance(card, source, prescale);
    return (OK);
}
